/*
 *     Mythtvfs: FUSE file system access to mythtv recordings.
 *
 *     Copyright (C) 2013 Mogens Kjr <mk@lemo.dk>
 *
 *             This program can be distributed under the terms of the GNU GPL.
 *
 *     The primary purpose of this program is to present the recordings in mythtv
 *     as a file system, ready for export to a DLNA client via coherence:
 *     http://coherence.beebits.net/
 *
 *     The file names in a folder is named with a four digit serial number, followed by a space
 *     and the description of the picture. The numbers are ordered according to the correct
 *     picture sequence in the album.
 *
 *     How to use it:
 *
 *     mkdir /tmp/mythtv
 *     ./mythtvfs -s /tmp/mythtv
 *
 *     /usr/bin/coherence --daemon --plugin=backend:FSStore,content:/tmp/mythtv,name:MythTV\ Recordings
 *
 *     To stop the service:
 *
 *     killall coherence
 *     fusermount -u /tmp/mythtv
 *
 *     You can also simply NFS export the /tmp/mythtv directory, making it accessible
 *     on OpenELEC on a Raspberry Pi.
 *
 *     The layout of the file system is:
 *
 *     channel/chanid progstart title
 *     date/progstart chanid title
 *     title/title chanid progstart
 */

#include <stdlib.h>
#define FUSE_USE_VERSION 26
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <mysql.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include "mythtvfs.h"
#include <stringprep.h>
#include <iconv.h>

#define MYSQL_UPDATE_INTERVAL 300
#define VIDEO_EXTENSION ".avi"

static time_t last_mysql_read = 0;
static int nrecordings=0;
static RECORDINGS *recordings=NULL;
static FILE *logfile;
static iconv_t latin2utf;

static char *movestring(char *s)
{
  int slen;
  char *p;

  if(s==NULL)
  {
    p = malloc(1);
    *p = '\0';
  }
  else
  {
    slen=strlen(s);
    p = malloc(slen+1);
    strcpy(p, s);
  }
  return p;
}

static char *movestringutf(char *s)
{
  int inbytes, outbytes, i, outlen, st;
  char *inbuf, *outbuf;
  char tmpbuf[8192];
  char *p;
  int iconv_err;

  /*
     The text in my recordings are encoded in Latin1,
     convert the text into UTF-8 here.
  */

  inbytes = strlen(s);
  inbuf = s;

  outbytes = sizeof(tmpbuf);
  outbuf = tmpbuf;

  st=iconv(latin2utf, &inbuf, &inbytes, &outbuf, &outbytes);
  
  if(inbytes == 0)
  {
    outlen = sizeof(tmpbuf)-outbytes;
    p = malloc(outlen+1);
    for(i=0; i<outlen; i++) p[i]=tmpbuf[i];
    p[outlen] = '\0';
    return p;
  }
  else
  {
    /* Convertion failed; just use the original text */
    return movestring(s);
  }
}

static void delete_recordings()
{
  int i;

  if(recordings==NULL) return;

  for(i=0; i<nrecordings; i++)
  {
    free(recordings[i].chanid);
    free(recordings[i].starttime);
    free(recordings[i].endtime);
    free(recordings[i].title);
    free(recordings[i].subtitle);
    free(recordings[i].description);
    free(recordings[i].filesize);
    free(recordings[i].basename);
    free(recordings[i].progstart);
    free(recordings[i].progend);
    free(recordings[i].channel_filename);
    free(recordings[i].date_filename);
    free(recordings[i].title_filename);
  }

  free(recordings);
  recordings=NULL;
  nrecordings=0;
}

static void cleanfilename(char *p)
{
  /* Remove illegal characters from file name */
  while(*p)
  {
    if(*p == '/') *p = '%';
    if(*p == '\r') *p = ' ';
    if(*p == '\n') *p = ' ';
    if(*p == ':') *p = '.';
    if(*p == '?') *p = 'x';
    p++;
  }
}

static void read_mythtv_data()
{
  int i,st;
  MYSQL mysql;
  MYSQL_RES *res1;
  MYSQL_ROW row1;
  char query[8192];
  char tmp[8192];

  if(time(NULL) - last_mysql_read < MYSQL_UPDATE_INTERVAL)
  {
#ifdef DEBUG
    fprintf(logfile, "read_mythtv_data: Data already valid: %u.\n", time(NULL));
#endif
    return;
  }

#ifdef DEBUG
  fprintf(logfile, "read_mythtv_data: Data outdated, update: %u\n", time(NULL));
#endif
  last_mysql_read = time(NULL);

  mysql_init(&mysql);
  mysql_options(&mysql,MYSQL_READ_DEFAULT_GROUP,"mythtvfs");
  if (!mysql_real_connect(&mysql,"localhost","mythtv","mythtv","mythconverg",0,NULL,0))
  {
    fprintf(logfile, "Failed to connect to database: Error: %s\n",
              mysql_error(&mysql));
  }

  /* Read all recordings */

  sprintf(query, "SELECT *,UNIX_TIMESTAMP(progstart) FROM recorded");

  st = mysql_query(&mysql, query);
  res1 = mysql_store_result(&mysql);

  delete_recordings();

  nrecordings = mysql_num_rows(res1);
  recordings = malloc(sizeof(*recordings)*nrecordings);

#ifdef DEBUG
  fprintf(logfile, "nrecordings: %d\n", nrecordings);
#endif

  i=0;

  while((row1 = mysql_fetch_row(res1))!=NULL)
  {
    recordings[i].chanid = movestring(row1[0]);
    recordings[i].starttime = movestring(row1[1]);
    recordings[i].endtime = movestring(row1[2]);
    recordings[i].title = movestringutf(row1[3]);
    recordings[i].subtitle = movestringutf(row1[4]);
    recordings[i].description = movestringutf(row1[5]);
    recordings[i].filesize = movestring(row1[21]);
    sscanf(row1[21], "%lld", &recordings[i].filesizeint);
    recordings[i].basename = movestring(row1[31]);
    recordings[i].progstart = movestring(row1[32]);
    recordings[i].progend = movestring(row1[33]);
    sscanf(row1[41], "%u", &recordings[i].progstartint);
    sprintf(tmp, "%s %s %s%s", recordings[i].chanid, recordings[i].progstart, recordings[i].title, VIDEO_EXTENSION);
    cleanfilename(tmp);
    recordings[i].channel_filename = movestring(tmp);
    sprintf(tmp, "%s %s %s%s", recordings[i].progstart, recordings[i].chanid, recordings[i].title, VIDEO_EXTENSION);
    cleanfilename(tmp);
    recordings[i].date_filename = movestring(tmp);
    sprintf(tmp, "%s %s %s%s", recordings[i].title, recordings[i].chanid, recordings[i].progstart, VIDEO_EXTENSION);
    cleanfilename(tmp);
    recordings[i].title_filename = movestring(tmp);
    i++;
  }
  mysql_free_result(res1);

#ifdef DEBUG2

  for(i=0;i<nrecordings;i++)
  {
    fprintf(logfile, "%4d %s %s %s %s %s %s %s %s %s %s %s %s %s %lld %u\n",
      i,
      recordings[i].chanid,
      recordings[i].starttime,
      recordings[i].endtime,
      recordings[i].title,
      recordings[i].subtitle,
      recordings[i].description,
      recordings[i].filesize,
      recordings[i].basename,
      recordings[i].progstart,
      recordings[i].progend,
      recordings[i].channel_filename,
      recordings[i].date_filename,
      recordings[i].title_filename,
      recordings[i].filesizeint,
      recordings[i].progstartint);
  }
  fflush(logfile);
#endif

  mysql_close(&mysql);

}

static void make_channel_path(char *path, int irecording)
{
 /* channel/chanid progstart title */
 sprintf(path, "/channel/%s", recordings[irecording].channel_filename);
}

static void make_date_path(char *path, int irecording)
{
 /* date/progstart chanid title */
 sprintf(path, "/date/%s", recordings[irecording].date_filename);
}

static void make_title_path(char *path, int irecording)
{
 /* title/title chanid progstart */
 sprintf(path, "/title/%s", recordings[irecording].title_filename);
}

static int find_path(char *path)
{
  /* Returns:

     0: Root folder
     1: channel folder
     2: date folder
     3: title folder
     -1:Everything else
  */
  int i;
  char dirpath[8192];
  int lastslash=-1;

  if(!strcmp(path, "/")) return 0;
  if(!strcmp(path, "/channel")) return 1;
  if(!strcmp(path, "/date")) return 2;
  if(!strcmp(path, "/title")) return 3;

  for(i=0; path[i]; i++)
  {
    dirpath[i] = path[i];
    if(path[i] == '/')
    {
      lastslash = i;
    }
  }

  if(lastslash == -1) return 0;
  dirpath[lastslash] = '\0';
  if(!strcmp(dirpath, "channel")) return 1;
  if(!strcmp(dirpath, "date")) return 2;
  if(!strcmp(dirpath, "title")) return 3;
  return -1;
}

static int find_recording(char *path)
{
  int i;
  char tmp[8192];

  for(i=0; i<nrecordings; i++)
  {
    make_channel_path(tmp, i);
    if(!strcmp(tmp, path)) return i;
    make_date_path(tmp, i);
    if(!strcmp(tmp, path)) return i;
    make_title_path(tmp, i);
    if(!strcmp(tmp, path)) return i;
  }
  return -1;
}



static int mythtvfs_getattr(const char *path, struct stat *stbuf)
{
  int res = -ENOENT;
  int irecording;

#ifdef DEBUG
  fprintf(logfile, "getattr: %s\n", path);
  fflush(logfile);
#endif

  memset(stbuf, 0, sizeof(struct stat));
  irecording = find_recording((char *) path);

  if(irecording < 0)
  {
    stbuf->st_mode = S_IFDIR | 0555;
    stbuf->st_nlink = 2;
    res = 0;
  }
  else
  {
    stbuf->st_mode = S_IFREG | 0444;
    stbuf->st_nlink = 1;
    stbuf->st_mtime = recordings[irecording].progstartint;
    stbuf->st_ctime = recordings[irecording].progstartint;
    stbuf->st_atime = recordings[irecording].progstartint;
    stbuf->st_size = recordings[irecording].filesizeint;
    stbuf->st_blksize = 512;
    stbuf->st_blocks = (recordings[irecording].filesizeint+511)/512;
#ifdef DEBUG
    fprintf(logfile, "getattr: %s size: %lld\n", path, stbuf->st_size);
    fflush(logfile);
#endif
    res = 0;
  }

  return res;
}


static int mythtvfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                         off_t offset, struct fuse_file_info *fi)
{
  int irecording,idir,i;
  (void) offset;
  (void) fi;

#ifdef DEBUG
  fprintf(logfile, "readdir: %s %ld\n", path, offset);
  fflush(logfile);
#endif
  read_mythtv_data();

  idir= find_path((char *) path);
  switch(idir)
  {
  case 0:
    /* root */
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    filler(buf, "channel", NULL, 0);
    filler(buf, "date", NULL, 0);
    filler(buf, "title", NULL, 0);
    break;
  case 1:
    /* channel */
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    for(i=0; i<nrecordings; i++)
      if(filler(buf, recordings[i].channel_filename, NULL, 0)!=0) return -ENOMEM;
    break;
  case 2:
    /* date */
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    for(i=0; i<nrecordings; i++)
      if(filler(buf, recordings[i].date_filename, NULL, 0)!=0) return -ENOMEM;
    break;
  case 3:
    /* title */
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    for(i=0; i<nrecordings; i++)
      if(filler(buf, recordings[i].title_filename, NULL, 0)!=0) return -ENOMEM;
    break;
  otherwise:
    return -ENOENT;
  }
  return 0;
}

static int mythtvfs_open(const char *path, struct fuse_file_info *fi)
{
  if((fi->flags & 3) != O_RDONLY)
    return -EACCES;

  return 0;
}

static int mythtvfs_read(const char *path, char *buf, size_t size, off_t offset,
                      struct fuse_file_info *fi)
{
  int irecording;
  int fh;
  off_t len;
  (void) fi;
  char filepath[8192];

#ifdef DEBUG2
  fprintf(logfile, "read: %s %d %lld\n", path, size, offset);
  fflush(logfile);
#endif

  irecording = find_recording((char *) path);
  if(irecording < 0) return -ENOENT;

  sprintf(filepath, "/var/video/%s", recordings[irecording].basename);

  fh=open(filepath, O_RDONLY);
  if(fh== -1) return -ENOENT;

  if(offset < recordings[irecording].filesizeint)
  {
    if(offset + size > recordings[irecording].filesizeint) size = recordings[irecording].filesizeint-offset;
    lseek(fh, offset, SEEK_SET);
    read(fh, buf, size);
  }
  else
    size=0;

  close(fh);

  return size;
}


static struct fuse_operations mythtvfs_oper = 
{
  .getattr = mythtvfs_getattr,
  .readdir = mythtvfs_readdir,
  .open	   = mythtvfs_open,
  .read	   = mythtvfs_read,
};

int main(int argc, char *argv[])
{
  latin2utf = iconv_open("UTF8", "LATIN1");
#ifdef DEBUG
  logfile = fopen("/tmp/mythtvfs.log", "w");
#endif
  read_mythtv_data();
  return fuse_main(argc, argv, &mythtvfs_oper, NULL);
}


